package de.hhu.phil.smt.lm;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Scanner;
import java.util.Set;

import de.hhu.phil.smt.SMT;
import de.hhu.phil.smt.misc.Pair;

/*
 * Author: kaeshammer@uni-duesseldorf.de
 * 
 * This class represents a bigram language model.
 */
public class BigramModel implements LanguageModel {
	
	private final int order = 2;
	
	// data structure for the vocabulary
	private Set<String> vocabulary;
	
	// data structures to hold counts for unigrams and bigrams
	private HashMap<String, Integer> uniCount;
	private HashMap<Pair<String, String>, Integer> biCount;
	
	/*
	 * Constructor reads the training data file which contains one
	 * sentence per line (that is tokenized by whitespace) and
	 * counts unigrams and bigrams.
	 */
	public BigramModel(String dataFile, Set<String> vocab) throws FileNotFoundException {
		vocabulary = vocab;
		
		uniCount = new HashMap<String, Integer>();
		biCount = new HashMap<Pair<String, String>, Integer>();
		
		// read in the training data und fill the data structures
		// uniCount and biCount
		Scanner input = new Scanner(new File(dataFile));
		String line = "";
		int c = 0;
		while (input.hasNextLine()) {
			line = input.nextLine();
			//System.out.println(line);
			c++;
			addToCounts(line);
		}
		
		input.close();
		System.out.println("The training data has " + c + " lines.");
	}
	
	@Override
	public int order() {
		return order;
	}
	
	// fill the uniCount and biCount data structures
	private void addToCounts(String line) {
		String lineWithStartEnd = SMT.addStartEnd(line);
		String[] tokens = SMT.tokenizeWS(lineWithStartEnd);
		
		String prev = null;
		for (String cur : tokens) {
			// low frequceny word -> unknown
			if (!vocabulary.contains(cur)) {
				cur = SMT.UNK;
			}
			
			// uniCount
			Integer c1 = uniCount.get(cur);
			if (c1 == null) {
				c1 = 0;
			}
			uniCount.put(cur, c1+1);
			
			// biCount
			if (prev != null) {
				Pair<String, String> bigram = new Pair<String, String>(prev, cur);
				Integer c2 = biCount.get(bigram);
				if (c2 == null) {
					c2 = 0;
				}
				biCount.put(bigram, c2+1);
			}	
			
			prev = cur;
		}		
	}

	// Calculate the probability P(w2|w1) using MLE and return it
	public double getProbMLE(String w1, String w2) {
		if (!vocabulary.contains(w1)) {
			w1 = SMT.UNK;
		}
		if (!vocabulary.contains(w2)) {
			w2 = SMT.UNK;
		}
		int c1 = 0;
		int c2 = 0;
		if (uniCount.containsKey(w1)) {
			c1 = uniCount.get(w1);
		}
		Pair<String, String> bigram = new Pair<String, String>(w1, w2);
		if (biCount.containsKey(bigram)) {
			c2 = biCount.get(bigram);
		}
		if (c1 == 0) { // should never happen
			throw new IllegalStateException("Unigram count is zero.");
		}
		return (double)c2/c1;
	}
	
	// Calculate the probability P(w2|w1) using MLE and add-one smoothing,
	// and return it
	public double getProbAddOneSmooth(String w1, String w2) {
		if (!vocabulary.contains(w1)) {
			w1 = SMT.UNK;
		}
		if (!vocabulary.contains(w2)) {
			w2 = SMT.UNK;
		}
		
		int c1 = 0;
		int c2 = 0;
		if (uniCount.containsKey(w1)) {
			c1 = uniCount.get(w1);
		}
		Pair<String, String> bigram = new Pair<String, String>(w1, w2);
		if (biCount.containsKey(bigram)) {
			c2 = biCount.get(bigram);
		}
		//System.out.println(bigram + " " + c2 + "    " + w1 + " " + c1);
		return (double)(c2+1)/(c1+vocabulary.size());
	}
	
	// P(w|h): MLE with Add-One-Smoothing
	@Override
	public double getProb(String h, String w) {
		return getProbAddOneSmooth(h, w);
	}
	
	// Calculate the probability that your bigram language model assigns
	// to the test sentence. Use the smoothed estimates.
	@Override
	public double score(String test) {
		String testWithStartEnd = SMT.addStartEnd(test);
		String[] tokens = SMT.tokenizeWS(testWithStartEnd);
		double res = 1;
		for (int i = 1; i < tokens.length; i++) {
			res *= getProbAddOneSmooth(tokens[i-1], tokens[i]);
		}
		return res;
	}
	
	// Calculate the probability that the language model assigns to the 
	// test sentences in the test file using the smoothed estimates.
	public double scoreFile(String dataFile) throws FileNotFoundException {
		Scanner input = new Scanner(new File(dataFile));
		String line = "";
		int c = 0;
		double s = 1;
		while (input.hasNextLine()) {
			line = input.nextLine();
			c++;
			s *= score(line);
			System.out.println(s);
		}
		System.out.println("The test data has " + c + " lines.");
		input.close();
		return s;
	}
	
	// log-probability log P(w2|w1) (MLE and add-one smoothing)
	public double getLogProbAddOneSmooth(String w1, String w2) {
		return Math.log(getProbAddOneSmooth(w1, w2));
	}
	

	// Calculate the log-probability that your bigram language model assigns
	// to the test sentence. Use the smoothed estimates.
	@Override
	public double logScore(String test) {
		String testWithStartEnd = SMT.addStartEnd(test);
		String[] tokens = SMT.tokenizeWS(testWithStartEnd);
		double res = 0;
		for (int i = 1; i < tokens.length; i++) {
			res += getLogProbAddOneSmooth(tokens[i-1], tokens[i]);
		}
		return res;
	}
	

	// Calculate the log-probability that the language model assigns to the 
	// test sentences in the test file using the smoothed estimates.
	@Override
	public double logScoreFile(String dataFile) throws FileNotFoundException {
		Scanner input = new Scanner(new File(dataFile));
		String line = "";
		int c = 0;
		double s = 0;
		while (input.hasNextLine()) {
			line = input.nextLine();
			c++;
			s += logScore(line);
			System.out.println(s);
		}
		System.out.println("The test data has " + c + " lines.");
		input.close();
		return s;
	}
	
	

}
