package de.hhu.phil.smt.lm;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Scanner;
import java.util.Set;

import de.hhu.phil.smt.SMT;

/*
 * Author: kaeshammer@uni-duesseldorf.de
 * 
 * This class represents a trigram language model.
 */
public class TrigramModel implements LanguageModel {
	
	private final int order = 3;
	
	// data structure for the vocabulary
	private Set<String> vocabulary;
	
	// data structures to hold counts for bigrams and trigrams
	private HashMap<String, Integer> biCount;
	private HashMap<String, Integer> triCount;

	/*
	 * Constructor reads the training data file which contains one
	 * sentence per line (that is tokenized by whitespace) and
	 * counts bigrams and trigrams.
	 */
	public TrigramModel(String dataFile, Set<String> vocab) throws FileNotFoundException {
		vocabulary = vocab;
		
		biCount = new HashMap<String, Integer>();
		triCount = new HashMap<String, Integer>();
		
		// read in the training data und fill the data structures
		// uniCount and biCount
		Scanner input = new Scanner(new File(dataFile));
		String line = "";
		int c = 0;
		while (input.hasNextLine()) {
			line = input.nextLine();
			//System.out.println(line);
			c++;
			addToCounts(line);
		}
		
		input.close();
		System.out.println("The training data has " + c + " lines.");
	}
	
	// fill the biCount and triCount data structures
	private void addToCounts(String line) {
		String lineWithStartEnd = SMT.addStartStartEnd(line);
		String[] tokens = SMT.tokenizeWS(lineWithStartEnd);
		
		String prev = null;
		String prevprev = null;
		for (String cur : tokens) {
			// low frequency word -> unknown
			if (!vocabulary.contains(cur)) {
				cur = SMT.UNK;
			}
			
			// biCount
			if (prev != null) {
				String bigram = prev + " " + cur;
				Integer c2 = biCount.get(bigram);
				if (c2 == null) {
					c2 = 0;
				}
				biCount.put(bigram, c2+1);
			}	
			
			// triCount
			if (prev != null && prevprev != null) {
				String trigram = prevprev + " " + prev + " " + cur;
				Integer c3 = triCount.get(trigram);
				if (c3 == null) {
					c3 = 0;
				}
				triCount.put(trigram, c3+1);
			}
			
			prevprev = prev;
			prev = cur;
		}		
	}
	
	@Override
	public int order() {
		return order;
	}

	@Override
	public double getProb(String h, String w) {
		String[] tokens = SMT.tokenizeWS(h);
		if (tokens.length != order-1) {
			throw new IllegalStateException("Parameter does not correspond to the order of the LM.");
			// In a perfect world, we would define our own exceptions...
		}
		return getProbAddOneSmooth(tokens[0], tokens[1], w);
	}
	
	// Calculate the probability P(w3|w1,w2) using MLE and add-one smoothing,
	// and return it
	public double getProbAddOneSmooth(String w1, String w2, String w3) {
		if (!vocabulary.contains(w1)) {
			w1 = SMT.UNK;
		}
		if (!vocabulary.contains(w2)) {
			w2 = SMT.UNK;
		}
		if (!vocabulary.contains(w3)) {
			w3 = SMT.UNK;
		}
		
		int c2 = 0;
		int c3 = 0;
		String bigram = w1 + " " + w2;
		String trigram = bigram + " " + w3;
		if (biCount.containsKey(bigram)) {
			c2 = biCount.get(bigram);
		}
		if (triCount.containsKey(trigram)) {
			c3 = triCount.get(trigram);
		}
		//System.out.println(trigram + " " + c3 + "    " + bigram + " " + c2);
		return (double)(c3+1)/(c2+vocabulary.size());
	}
	
	
	// log-probability log P(w3|w1,w2) using MLE and add-one smoothing
	public double getLogProbAddOneSmooth(String w1, String w2,String w3) {
		return Math.log(getProbAddOneSmooth(w1, w2, w3));
	}

	// Calculate the probability that your trigram language model assigns
	// to the test sentence. Use the smoothed estimates.
	@Override
	public double score(String test) {
		String testWithStartEnd = SMT.addStartStartEnd(test);
		String[] tokens = SMT.tokenizeWS(testWithStartEnd);
		double res = 1;
		for (int i = 2; i < tokens.length; i++) {
			res *= getProbAddOneSmooth(tokens[i-2], tokens[i-1], tokens[i]);
		}
		return res;
	}
	
	// Calculate the log-probability that your trigram language model assigns
	// to the test sentence. Use the smoothed estimates.
	@Override
	public double logScore(String test) {
		String testWithStartEnd = SMT.addStartStartEnd(test);
		String[] tokens = SMT.tokenizeWS(testWithStartEnd);
		double res = 0;
		for (int i = 2; i < tokens.length; i++) {
			res += getLogProbAddOneSmooth(tokens[i-2], tokens[i-1], tokens[i]);
		}
		return res;
	}


	// Calculate the log probability that the language model assigns to the 
	// test sentences in the test file using the smoothed estimates.
	@Override
	public double logScoreFile(String dataFile) throws FileNotFoundException {
		Scanner input = new Scanner(new File(dataFile));
		String line = "";
		int c = 0;
		double s = 0;
		while (input.hasNextLine()) {
			line = input.nextLine();
			c++;
			s += logScore(line);
			System.out.println(s);
		}
		System.out.println("The test data has " + c + " lines.");
		input.close();
		return s;
	}


}
